# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import exceptions
from . import session_events
from ._about import __version__
from ._enums import AckStatus
from ._enums import CompressionAlgorithmType
from ._enums import PropertyType
from ._messages import Ack
from ._messages import Message
from ._messages import MessageHandle
from ._monitors import BasicHealthMonitor
from ._session import QueueOptions
from ._session import Session
from ._session import SessionOptions
from ._timeouts import Timeouts
from ._typing import PropertyTypeDict
from ._typing import PropertyValueDict
from .exceptions import Error

__all__ = [
    "Ack",
    "AckStatus",
    "BasicHealthMonitor",
    "CompressionAlgorithmType",
    "Error",
    "PropertyType",
    "PropertyTypeDict",
    "PropertyValueDict",
    "QueueOptions",
    "Message",
    "MessageHandle",
    "Session",
    "SessionOptions",
    "Timeouts",
    "__version__",
    "exceptions",
    "session_events",
]



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Optional


class Timeouts:
    """A value semantic type representing session timeouts.

    Each option can be set either by passing it as a keyword argument when
    constructing a *Timeouts* instance, or by setting it as an attribute on
    a constructed instance.

    The default for every option is `None`. When constructing a `Session`,
    either directly or using `SessionOptions`, options set to `None` are given
    reasonable default values.

    Args:
        connect_timeout:
            The maximum number of seconds to wait for connection requests on
            this session.
        disconnect_timeout:
            The maximum number of seconds to wait for disconnection requests
            on this session.
        open_queue_timeout:
            The maximum number of seconds to wait for open queue requests on
            this session.
        configure_queue_timeout:
            The maximum number of seconds to wait for configure queue requests
            on this session.
        close_queue_timeout:
            The maximum number of seconds to wait for close queue requests on
            this session.
    """

    def __init__(
        self,
        connect_timeout: Optional[float] = None,
        disconnect_timeout: Optional[float] = None,
        open_queue_timeout: Optional[float] = None,
        configure_queue_timeout: Optional[float] = None,
        close_queue_timeout: Optional[float] = None,
    ) -> None:
        self.connect_timeout = connect_timeout
        self.disconnect_timeout = disconnect_timeout
        self.open_queue_timeout = open_queue_timeout
        self.configure_queue_timeout = configure_queue_timeout
        self.close_queue_timeout = close_queue_timeout

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Timeouts):
            return False
        return (
            self.connect_timeout == other.connect_timeout
            and self.disconnect_timeout == other.disconnect_timeout
            and self.open_queue_timeout == other.open_queue_timeout
            and self.configure_queue_timeout == other.configure_queue_timeout
            and self.close_queue_timeout == other.close_queue_timeout
        )

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __repr__(self) -> str:
        attrs = (
            "connect_timeout",
            "disconnect_timeout",
            "open_queue_timeout",
            "configure_queue_timeout",
            "close_queue_timeout",
        )

        params = []
        for attr in attrs:
            value = getattr(self, attr)
            if value is not None:
                params.append(f"{attr}={value!r}")

        return f"Timeouts({', '.join(params)})"



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The `blazingmq.testing` module provides testing utilities.

Nothing in this module should ever be used outside of the context of performing
integration tests on your subscribers and publishers.
"""

from ._monitors import BasicHealthMonitor

HostHealth = BasicHealthMonitor
"""This is an alias of `.BasicHealthMonitor`.

This was originally exposed only for testing purposes, but is now supported for
general use. The original name has been retained for backwards compatibility,
but new code should prefer to use `blazingmq.BasicHealthMonitor` directly.
"""



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import atexit
from functools import partial
import logging
import weakref

from bsl cimport optional
from bsl cimport pair
from bsl cimport shared_ptr
from bsl.bsls cimport TimeInterval
from cpython.ceval cimport PyEval_InitThreads
from libcpp cimport bool as cppbool

from bmq.bmqa cimport ManualHostHealthMonitor
from bmq.bmqt cimport AckResult
from bmq.bmqt cimport CompressionAlgorithmType
from bmq.bmqt cimport HostHealthState
from bmq.bmqt cimport PropertyType
from bmq.bmqt cimport SessionEventEnum
from bmq.bmqt cimport k_DEFAULT_CONSUMER_PRIORITY
from bmq.bmqt cimport k_DEFAULT_MAX_UNCONFIRMED_BYTES
from bmq.bmqt cimport k_DEFAULT_MAX_UNCONFIRMED_MESSAGES
from bmq.bmqt cimport k_DEFAULT_SUSPENDS_ON_BAD_HOST_HEALTH
from pybmq cimport BallUtil
from pybmq cimport Session as NativeSession

from typing import Optional

from . import _callbacks
from . import _enums
from . import _messages
from . import _script_name
from . import _timeouts
from . import session_events
from .exceptions import BrokerTimeoutError
from .exceptions import Error

LOGGER = logging.getLogger(__name__)

BallUtil.initBallSingleton(_log_callback, "BlazingMQ C++ log observer")
atexit.register(BallUtil.shutDownBallSingleton)

cdef _log_callback(const char *name,
                   int level,
                   const char *filename,
                   int line,
                   const char *msg) noexcept:
    if not LOGGER.isEnabledFor(level):
        return
    rec = LOGGER.makeRecord(name, level, filename, line, msg, (), None)
    LOGGER.handle(rec)


SESSION_EVENT_TYPE_MAPPING = {
    SessionEventEnum.e_CONNECTED: session_events.Connected,
    SessionEventEnum.e_DISCONNECTED: session_events.Disconnected,
    SessionEventEnum.e_CONNECTION_LOST: session_events.ConnectionLost,
    SessionEventEnum.e_RECONNECTED: session_events.Reconnected,
    SessionEventEnum.e_STATE_RESTORED: session_events.StateRestored,
    SessionEventEnum.e_CONNECTION_TIMEOUT: session_events.ConnectionTimeout,
    SessionEventEnum.e_QUEUE_REOPEN_RESULT: session_events.QueueReopened,
    SessionEventEnum.e_SLOWCONSUMER_NORMAL: session_events.SlowConsumerNormal,
    SessionEventEnum.e_SLOWCONSUMER_HIGHWATERMARK: session_events.SlowConsumerHighWaterMark,
    SessionEventEnum.e_ERROR: session_events.Error,
    SessionEventEnum.e_HOST_UNHEALTHY: session_events.HostUnhealthy,
    SessionEventEnum.e_HOST_HEALTH_RESTORED: session_events.HostHealthRestored,
    SessionEventEnum.e_QUEUE_SUSPENDED: session_events.QueueSuspended,
    SessionEventEnum.e_QUEUE_RESUMED: session_events.QueueResumed,
}


ACK_STATUS_MAPPING = {
    AckResult.e_SUCCESS: _messages.AckStatus.SUCCESS,
    AckResult.e_UNKNOWN: _messages.AckStatus.UNKNOWN,
    AckResult.e_TIMEOUT: _messages.AckStatus.TIMEOUT,
    AckResult.e_NOT_CONNECTED: _messages.AckStatus.NOT_CONNECTED,
    AckResult.e_CANCELED: _messages.AckStatus.CANCELED,
    AckResult.e_NOT_SUPPORTED: _messages.AckStatus.NOT_SUPPORTED,
    AckResult.e_REFUSED: _messages.AckStatus.REFUSED,
    AckResult.e_INVALID_ARGUMENT: _messages.AckStatus.INVALID_ARGUMENT,
    AckResult.e_NOT_READY: _messages.AckStatus.NOT_READY,
    AckResult.e_LIMIT_BYTES: _messages.AckStatus.LIMIT_BYTES,
    AckResult.e_LIMIT_MESSAGES: _messages.AckStatus.LIMIT_MESSAGES,
    AckResult.e_STORAGE_FAILURE: _messages.AckStatus.STORAGE_FAILURE,
}

PROPERTY_TYPES_TO_PY_MAPPING = {
    PropertyType.e_BOOL: _enums.PropertyType.BOOL,
    PropertyType.e_CHAR: _enums.PropertyType.CHAR,
    PropertyType.e_SHORT: _enums.PropertyType.SHORT,
    PropertyType.e_INT32: _enums.PropertyType.INT32,
    PropertyType.e_INT64: _enums.PropertyType.INT64,
    PropertyType.e_STRING: _enums.PropertyType.STRING,
    PropertyType.e_BINARY: _enums.PropertyType.BINARY,
}

PROPERTY_TYPES_FROM_PY_MAPPING = {v: k for k, v in PROPERTY_TYPES_TO_PY_MAPPING.items()}

COMPRESSION_ALGO_FROM_PY_MAPPING = {
    _enums.CompressionAlgorithmType.NONE: CompressionAlgorithmType.e_NONE,
    _enums.CompressionAlgorithmType.ZLIB: CompressionAlgorithmType.e_ZLIB,
}

DEFAULT_MAX_UNCONFIRMED_MESSAGES = k_DEFAULT_MAX_UNCONFIRMED_MESSAGES
DEFAULT_MAX_UNCONFIRMED_BYTES = k_DEFAULT_MAX_UNCONFIRMED_BYTES
DEFAULT_CONSUMER_PRIORITY = k_DEFAULT_CONSUMER_PRIORITY
DEFAULT_SUSPENDS_ON_BAD_HOST_HEALTH = k_DEFAULT_SUSPENDS_ON_BAD_HOST_HEALTH

cdef TimeInterval create_time_interval(timeout: Optional[int|float]=None):
    if timeout is not None:
        return TimeInterval(timeout)
    return TimeInterval()


cdef ensure_stop_session_impl(weakref_ext_session):
    session = weakref_ext_session()
    if session is not None:
        (<Session?>session)._session.stop(True)


def ensure_stop_session(weakref_ext_session):
    ensure_stop_session_impl(weakref_ext_session)


cdef class FakeHostHealthMonitor:
    cdef shared_ptr[ManualHostHealthMonitor] _monitor

    def __cinit__(self):
        with nogil:
            self._monitor = shared_ptr[ManualHostHealthMonitor](
                new ManualHostHealthMonitor(HostHealthState.e_HEALTHY)
            )

    def set_healthy(self):
        with nogil:
            self._monitor.get().setState(HostHealthState.e_HEALTHY)

    def set_unhealthy(self):
        with nogil:
            self._monitor.get().setState(HostHealthState.e_UNHEALTHY)


cdef class Session:
    cdef object __weakref__
    cdef NativeSession* _session
    cdef readonly object monitor_host_health

    def __cinit__(
        self,
        on_session_event not None,
        *,
        on_message=None,
        broker not None: bytes = b'tcp://localhost:30114',
        message_compression_algorithm not None=_enums.CompressionAlgorithmType.NONE,
        num_processing_threads: Optional[int] = None,
        blob_buffer_size: Optional[int] = None,
        channel_high_watermark: Optional[int] = None,
        event_queue_watermarks: Optional[tuple[int,int]] = None,
        stats_dump_interval: Optional[int|float] = None,
        timeouts: _timeouts.Timeouts = _timeouts.Timeouts(),
        monitor_host_health: bool = False,
        fake_host_health_monitor: FakeHostHealthMonitor = None,
        _mock: Optional[object] = None,
    ) -> None:
        cdef shared_ptr[ManualHostHealthMonitor] fake_host_health_monitor_sp
        cdef optional[int] c_num_processing_threads
        cdef optional[int] c_blob_buffer_size
        cdef optional[int] c_channel_high_watermark
        cdef optional[pair[int,int]] c_event_queue_watermarks
        cdef TimeInterval c_stats_dump_interval = create_time_interval(stats_dump_interval)
        cdef TimeInterval c_connect_timeout = create_time_interval(timeouts.connect_timeout)
        cdef TimeInterval c_disconnect_timeout = create_time_interval(timeouts.disconnect_timeout)
        cdef TimeInterval c_open_queue_timeout = create_time_interval(timeouts.open_queue_timeout)
        cdef TimeInterval c_configure_queue_timeout = create_time_interval(timeouts.configure_queue_timeout)
        cdef TimeInterval c_close_queue_timeout = create_time_interval(timeouts.close_queue_timeout)

        PyEval_InitThreads()

        if num_processing_threads is not None:
            c_num_processing_threads = optional[int](num_processing_threads)
        if blob_buffer_size is not None:
            c_blob_buffer_size = optional[int](blob_buffer_size)
        if channel_high_watermark is not None:
            c_channel_high_watermark = optional[int](channel_high_watermark)
        if event_queue_watermarks is not None:
            c_event_queue_watermarks = optional[pair[int,int]](
                pair[int,int](event_queue_watermarks[0], event_queue_watermarks[1]))

        self.monitor_host_health = monitor_host_health

        if fake_host_health_monitor:
            fake_host_health_monitor_sp = fake_host_health_monitor._monitor

        session_cb = partial(_callbacks.on_session_event,
                             on_session_event,
                             SESSION_EVENT_TYPE_MAPPING)
        if on_message is None:
            message_cb = partial(_callbacks.on_message_create_interface_error, on_session_event)
        else:
            message_cb = partial(
                _callbacks.on_message,
                on_message,
                weakref.ref(self),
                PROPERTY_TYPES_TO_PY_MAPPING,
            )
        ack_cb = partial(_callbacks.on_ack, ACK_STATUS_MAPPING)
        cdef char *c_broker_uri = broker
        script_name = _script_name.get_script_name()
        cdef char *c_script_name = script_name
        self._session = new NativeSession(
            session_cb,
            message_cb,
            ack_cb,
            c_broker_uri,
            c_script_name,
            COMPRESSION_ALGO_FROM_PY_MAPPING[message_compression_algorithm],
            c_num_processing_threads,
            c_blob_buffer_size,
            c_channel_high_watermark,
            c_event_queue_watermarks,
            c_stats_dump_interval,
            c_connect_timeout,
            c_disconnect_timeout,
            c_open_queue_timeout,
            c_configure_queue_timeout,
            c_close_queue_timeout,
            monitor_host_health,
            fake_host_health_monitor_sp,
            Error,
            BrokerTimeoutError,
            _mock)
        self._session.start(c_connect_timeout)
        atexit.register(ensure_stop_session_impl, weakref.ref(self))

    def stop(self) -> None:
        self._session.stop(False)

    def open_queue_sync(self,
                        queue_uri not None: bytes,
                        *,
                        read: bool,
                        write: bool,
                        consumer_priority: Optional[int] = None,
                        max_unconfirmed_messages: Optional[int] = None,
                        max_unconfirmed_bytes: Optional[int] = None,
                        suspends_on_bad_host_health: Optional[bool] = None,
                        timeout: Optional[int|float] = None) -> None:
        cdef optional[int] c_consumer_priority
        cdef optional[int] c_max_unconfirmed_messages
        cdef optional[int] c_max_unconfirmed_bytes
        cdef optional[cppbool] c_suspends_on_bad_host_health
        cdef TimeInterval c_timeout = create_time_interval(timeout)

        if consumer_priority is not None:
            c_consumer_priority = optional[int](consumer_priority)

        if max_unconfirmed_messages is not None:
            c_max_unconfirmed_messages = optional[int](max_unconfirmed_messages)

        if max_unconfirmed_bytes is not None:
            c_max_unconfirmed_bytes = optional[int](max_unconfirmed_bytes)

        if suspends_on_bad_host_health is not None:
            c_suspends_on_bad_host_health = optional[cppbool](suspends_on_bad_host_health)

        self._session.open_queue_sync(queue_uri,
                                      read,
                                      write,
                                      c_consumer_priority,
                                      c_max_unconfirmed_messages,
                                      c_max_unconfirmed_bytes,
                                      c_suspends_on_bad_host_health,
                                      c_timeout)

    def configure_queue_sync(self,
                             queue_uri not None: bytes,
                             *,
                             consumer_priority: Optional[int] = None,
                             max_unconfirmed_messages: Optional[int] = None,
                             max_unconfirmed_bytes: Optional[int] = None,
                             suspends_on_bad_host_health: Optional[bool] = None,
                             timeout: Optional[int|float] = None) -> None:
        cdef optional[int] c_consumer_priority
        cdef optional[int] c_max_unconfirmed_messages
        cdef optional[int] c_max_unconfirmed_bytes
        cdef optional[cppbool] c_suspends_on_bad_host_health
        cdef TimeInterval c_timeout = create_time_interval(timeout)

        if consumer_priority is not None:
            c_consumer_priority = optional[int](consumer_priority)

        if max_unconfirmed_messages is not None:
            c_max_unconfirmed_messages = optional[int](max_unconfirmed_messages)

        if max_unconfirmed_bytes is not None:
            c_max_unconfirmed_bytes = optional[int](max_unconfirmed_bytes)

        if suspends_on_bad_host_health is not None:
            c_suspends_on_bad_host_health = optional[cppbool](suspends_on_bad_host_health)

        self._session.configure_queue_sync(queue_uri,
                                           c_consumer_priority,
                                           c_max_unconfirmed_messages,
                                           c_max_unconfirmed_bytes,
                                           c_suspends_on_bad_host_health,
                                           c_timeout)

    def close_queue_sync(self,
                         queue_uri not None: bytes,
                         timeout: Optional[int|float] = None) -> None:
        cdef TimeInterval c_timeout = create_time_interval(timeout)
        self._session.close_queue_sync(queue_uri, c_timeout)

    def get_queue_options(self,
                          queue_uri not None: bytes) -> object:
        return self._session.get_queue_options(queue_uri)

    def post(self,
             queue_uri not None: bytes,
             payload not None: bytes,
             properties=None,
             on_ack=None) -> None:
        self._session.post(queue_uri, payload, len(payload), properties, on_ack)

    def confirm(self, message not None) -> None:
        self._session.confirm(message.queue_uri, message.guid, len(message.guid))

    def __dealloc__(self) -> None:
        if self._session:
            try:
                self._session.stop(True)
            finally:
                del self._session



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

cdef extern from 'bdlt_datetime.h' namespace 'BloombergLP::bdlt' nogil:
    cdef cppclass Datetime:
        Datetime()
        Datetime(int, int, int, int, int, int, int, int)
        pass


cdef extern from 'bdlt_date.h' namespace 'BloombergLP::bdlt' nogil:

    cdef cppclass Date:
        # CREATORS
        Date()
        Date(int, int, int)

        # MANIPULATORS
        void setYearMonthDay(int year, int month, int day)


cdef extern from 'bdlt_datetz.h' namespace 'BloombergLP::bdlt' nogil:

    cdef cppclass DateTz:
        # CREATORS
        DateTz()

        # MANIPULATORS
        # void setYearMonthDay(int year, int month, int day)


cdef extern from 'bdlt_time.h' namespace 'BloombergLP::bdlt' nogil:

    cdef cppclass Time:
        # CREATORS
        Time()
        Time(int, int, int, int, int)

        # MANIPULATORS
        # Note that we elide the optional parameters to make Cython happy.
        void setTime(int hour, int minute, int second, int millisecond, int microsecond)


cdef extern from 'bdlt_timetz.h' namespace 'BloombergLP::bdlt' nogil:

    cdef cppclass TimeTz:
        # CREATORS
        TimeTz()


cdef extern from 'bdlt_datetimetz.h' namespace 'BloombergLP::bdlt' nogil:

    cdef cppclass DatetimeTz:
        # CREATORS
        DatetimeTz()

cdef extern from 'bdlt_datetimeinterval.h' namespace 'BloombergLP::bdlt' nogil:

    cdef cppclass DatetimeInterval:
        # CREATORS
        DatetimeInterval()
        DatetimeInterval(int, unsigned long long, unsigned long long, unsigned long long,
                         unsigned long long, unsigned long long)



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

from . import _six as six
from ._enums import CompressionAlgorithmType
from ._enums import PropertyType
from ._ext import DEFAULT_CONSUMER_PRIORITY
from ._ext import DEFAULT_MAX_UNCONFIRMED_BYTES
from ._ext import DEFAULT_MAX_UNCONFIRMED_MESSAGES
from ._ext import DEFAULT_SUSPENDS_ON_BAD_HOST_HEALTH
from ._ext import PROPERTY_TYPES_FROM_PY_MAPPING
from ._ext import Session as ExtSession
from ._messages import Ack
from ._messages import Message
from ._messages import MessageHandle
from ._monitors import BasicHealthMonitor
from ._timeouts import Timeouts
from ._typing import PropertyTypeDict
from ._typing import PropertyValueDict
from ._typing import PropertyValueType
from .exceptions import Error
from .session_events import SessionEvent


class DefaultTimeoutType(float):
    def __repr__(self) -> str:
        return "..."


def DefaultMonitor() -> Union[BasicHealthMonitor, None]:
    return None


DEFAULT_TIMEOUT = DefaultTimeoutType()
KNOWN_MONITORS = ("blazingmq.BasicHealthMonitor",)


def _validate_timeouts(timeouts: Timeouts) -> Timeouts:
    """Validate a `.Timeouts` instance for use by the Cython layer.

    If any of the timeouts contained within the `.Timeouts` instance are the
    `DEFAULT_TIMEOUT` sentinel or `None`, return `None`.  Otherwise, validate
    that it is within the range accepted by `bsls::TimeInterval` and return it.
    """
    return Timeouts(
        connect_timeout=_convert_timeout(timeouts.connect_timeout),
        disconnect_timeout=_convert_timeout(timeouts.disconnect_timeout),
        open_queue_timeout=_convert_timeout(timeouts.open_queue_timeout),
        configure_queue_timeout=_convert_timeout(timeouts.configure_queue_timeout),
        close_queue_timeout=_convert_timeout(timeouts.close_queue_timeout),
    )


def _convert_timeout(timeout: Optional[float]) -> Optional[float]:
    """Convert the timeout for use by the Cython layer.

    If it is the DEFAULT_TIMEOUT sentinel or None, return None.  Otherwise,
    validate that it is within the range accepted by bsls::TimeInterval and
    return it.
    """
    if timeout is DEFAULT_TIMEOUT or timeout is None:
        return None
    elif 0.0 < timeout < 2**63:
        return timeout
    raise ValueError(f"timeout must be greater than 0.0, was {timeout}")


def _convert_stats_dump_interval(interval: Optional[float]) -> Optional[float]:
    """Convert the stats dump interval for use by the Cython layer.

    If is None, return None.  Otherwise, validate that it is within the range
    accepted by bsls::TimeInterval and return it.
    """
    if interval is None:
        return interval
    if 0.0 <= interval < 2**63:
        return interval
    raise ValueError(f"stats_dump_interval must be nonnegative, was {interval}")


def _collect_properties_and_types(
    properties: Optional[PropertyValueDict],
    property_type_overrides: Optional[PropertyTypeDict],
) -> Dict[bytes, Tuple[Union[int, bytes], int]]:
    property_val_by_name: Dict[bytes, PropertyValueType] = {}
    property_type_by_name: Dict[bytes, PropertyType] = {}

    if properties:
        for name, val in properties.items():
            if isinstance(val, bool):
                default_type = PropertyType.BOOL
            elif isinstance(val, int):
                default_type = PropertyType.INT64
            elif isinstance(val, str):
                default_type = PropertyType.STRING
            elif isinstance(val, bytes):
                default_type = PropertyType.BINARY
            else:
                raise Error(
                    "Property values of type %r are not supported" % type(val).__name__
                )

            name_bytes = six.ensure_binary(name)
            property_val_by_name[name_bytes] = val
            property_type_by_name[name_bytes] = default_type

    if property_type_overrides:
        for name, override_type in property_type_overrides.items():
            name_bytes = six.ensure_binary(name)
            if name_bytes not in property_type_by_name:
                raise Error("Received override for non-existent property %r" % name)
            property_type_by_name[name_bytes] = override_type

    merged: Dict[bytes, Tuple[Union[int, bytes], int]] = {}
    for name_bytes, val in property_val_by_name.items():
        property_type = property_type_by_name[name_bytes]
        type_code = PROPERTY_TYPES_FROM_PY_MAPPING[property_type]
        if property_type is PropertyType.STRING and isinstance(val, str):
            merged[name_bytes] = (val.encode("utf-8"), type_code)
        else:
            merged[name_bytes] = (val, type_code)  # type: ignore[assignment]
            # mypy warns because `merged` can still contain Text if a Text value was
            # passed for a non-STRING field.  If so, the extension gives a nice error.

    return merged


class QueueOptions:
    """A value semantic type representing the settings for a queue.

    Each option can be set either by passing it as a keyword argument when
    constructing a *QueueOptions* instance, or by setting it as an attribute on
    a constructed instance.

    The default for every option is `None`. When calling `.configure_queue`,
    options set to `None` are not changed from their current setting. When
    calling `.open_queue`, options set to `None` are given default values.
    These default values are accessible as class attributes on the
    *QueueOptions* class.

    Args:
        max_unconfirmed_messages:
            The maximum number of messages that can be delivered to the queue
            without confirmation. If 0, no limit is imposed.  This limit can be
            reached if the queue receives messages faster than it confirms
            received messages. Once this limit is reached, at least one message
            must be confirmed before the queue will receive any more messages.
        max_unconfirmed_bytes:
            The maximum number of bytes that can be delivered to the queue
            without confirmation. If 0, no limit is imposed.  Like
            *max_unconfirmed_messages*, this limit can be reached if incoming
            messages are queued up waiting for already delivered messages to be
            confirmed.
        consumer_priority:
            The precedence of this consumer compared to other consumers of the
            same queue. The consumer with the highest priority receives the
            messages. If multiple consumers share the highest priority,
            messages are delivered to them in a round-robin fashion.
        suspends_on_bad_host_health:
            Whether or not this queue should suspend operation while the host
            machine is unhealthy. While operation is suspended, a queue opened
            with ``read=True`` will not receive messages and a queue opened
            with ``write=True`` will raise if you try to `.post` a message.
            By default, queues are not sensitive to the host's health.
    """

    DEFAULT_MAX_UNCONFIRMED_MESSAGES = DEFAULT_MAX_UNCONFIRMED_MESSAGES
    """The *max_unconfirmed_messages* used by `.open_queue` by default."""

    DEFAULT_MAX_UNCONFIRMED_BYTES = DEFAULT_MAX_UNCONFIRMED_BYTES
    """The *max_unconfirmed_bytes* used by `.open_queue` by default."""

    DEFAULT_CONSUMER_PRIORITY = DEFAULT_CONSUMER_PRIORITY
    """The *consumer_priority* used by `.open_queue` by default."""

    DEFAULT_SUSPENDS_ON_BAD_HOST_HEALTH = DEFAULT_SUSPENDS_ON_BAD_HOST_HEALTH
    """The *suspends_on_bad_host_health* used by `.open_queue` by default."""

    def __init__(
        self,
        max_unconfirmed_messages: Optional[int] = None,
        max_unconfirmed_bytes: Optional[int] = None,
        consumer_priority: Optional[int] = None,
        suspends_on_bad_host_health: Optional[bool] = None,
    ) -> None:
        self.max_unconfirmed_messages = max_unconfirmed_messages
        self.max_unconfirmed_bytes = max_unconfirmed_bytes
        self.consumer_priority = consumer_priority
        self.suspends_on_bad_host_health = suspends_on_bad_host_health

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, QueueOptions):
            return False
        return (
            self.max_unconfirmed_messages == other.max_unconfirmed_messages
            and self.max_unconfirmed_bytes == other.max_unconfirmed_bytes
            and self.consumer_priority == other.consumer_priority
            and self.suspends_on_bad_host_health == other.suspends_on_bad_host_health
        )

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __repr__(self) -> str:
        attrs = (
            "max_unconfirmed_messages",
            "max_unconfirmed_bytes",
            "consumer_priority",
            "suspends_on_bad_host_health",
        )

        params = []
        for attr in attrs:
            value = getattr(self, attr)
            if value is not None:
                params.append(f"{attr}={value!r}")

        return f"QueueOptions({', '.join(params)})"


class SessionOptions:
    """A value semantic type representing session options.

    Each option can be set either by passing it as a keyword argument when
    constructing a *SessionOptions* instance, or by setting it as an attribute
    on a constructed instance.

    The default for every option is `None`. When constructing a `Session`,
    options set to `None` are given reasonable default values.

    Args:
        message_compression_algorithm:
            The type of compression to apply to messages being posted via the
            session this object is configuring.
        timeouts:
            The maximum number of seconds to wait for requests for each
            operation on this session.  If not provided, reasonable defaults
            are used.
        host_health_monitor:
            A `.BasicHealthMonitor` is used by default, so your tests can
            control whether the session sees the machine as healthy or not by
            calling `.set_healthy` and `.set_unhealthy` on that instance.  If
            you instead pass `None`, the session will always see the machine as
            healthy, `.HostUnhealthy` and `.HostHealthRestored` events with
            never be emitted, and the *suspends_on_bad_host_health* option of
            `QueueOptions` cannot be used.
        num_processing_threads:
            The number of threads for the SDK to use for processing events.
            This defaults to 1.
        blob_buffer_size:
            The size (in bytes) of the blob buffers to use.  This defaults to
            4k.
        channel_high_watermark:
            The size (in bytes) to use for the write cache high watermark on
            the channel.  The default value is 128MB.  Note that BlazingMQ
            reserves 4MB of this value for control messages, so the actual
            watermark for data published is ``channel_high_watermark - 4MB``.
        event_queue_watermarks:
            A tuple containing the low and high notification watermark
            thresholds for the buffer containing all incoming messages from the
            broker, respectively.  A warning `.SlowConsumerHighWaterMark` is
            emitted when the buffer reaches the high watermark value, and a
            notification `.SlowConsumerNormal` is emitted when the buffer is
            back to the low watermark.
        stats_dump_interval:
            The interval (in seconds) at which to dump stats into the logs.  If
            0, disable the recurring dump of stats (final stats are always
            dumped at the end of the session).  The default is 5min; the value
            must be a multiple of 30s, in the range ``[0s - 60min]``.
    """

    def __init__(
        self,
        message_compression_algorithm: Optional[CompressionAlgorithmType] = None,
        timeouts: Optional[Timeouts] = None,
        host_health_monitor: Union[BasicHealthMonitor, None] = (DefaultMonitor()),
        num_processing_threads: Optional[int] = None,
        blob_buffer_size: Optional[int] = None,
        channel_high_watermark: Optional[int] = None,
        event_queue_watermarks: Optional[tuple[int, int]] = None,
        stats_dump_interval: Optional[float] = None,
    ) -> None:
        self.message_compression_algorithm = message_compression_algorithm
        self.timeouts = timeouts
        self.host_health_monitor = host_health_monitor
        self.num_processing_threads = num_processing_threads
        self.blob_buffer_size = blob_buffer_size
        self.channel_high_watermark = channel_high_watermark
        self.event_queue_watermarks = event_queue_watermarks
        self.stats_dump_interval = stats_dump_interval

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SessionOptions):
            return False
        return (
            self.message_compression_algorithm == other.message_compression_algorithm
            and self.timeouts == other.timeouts
            and self.host_health_monitor == other.host_health_monitor
            and self.num_processing_threads == other.num_processing_threads
            and self.blob_buffer_size == other.blob_buffer_size
            and self.channel_high_watermark == other.channel_high_watermark
            and self.event_queue_watermarks == other.event_queue_watermarks
            and self.stats_dump_interval == other.stats_dump_interval
        )

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __repr__(self) -> str:
        attrs = (
            "message_compression_algorithm",
            "timeouts",
            "host_health_monitor",
            "num_processing_threads",
            "blob_buffer_size",
            "channel_high_watermark",
            "event_queue_watermarks",
            "stats_dump_interval",
        )

        params = []
        for attr in attrs:
            value = getattr(self, attr)
            if value is not None:
                params.append(f"{attr}={value!r}")

        return f"SessionOptions({', '.join(params)})"


class Session:
    """Represents a connection with the BlazingMQ broker.

    The session represents a connection to the broker.
    This object can be manipulated to modify the state of the application from
    the point of view of the broker, including opening queues, and starting or
    stopping the connection with the broker.

    Args:
        on_session_event: a required callback to process `.SessionEvent` events
            received by the session.
        on_message: an optional callback to process `Message` objects received
            by the session.
        broker: TCP address of the broker (default: 'tcp://localhost:30114').
            If the environment variable ``BMQ_BROKER_URI`` is set, its value
            will override whatever broker address is passed via this argument.
        message_compression_algorithm: the type of compression to apply to messages
            being posted via this session object.
        timeout: maximum number of seconds to wait for requests on this
            session.  If not provided, reasonable defaults are used.  This
            argument may either be a simple ``float``, which sets the same
            timeout for each operation, or an instance of the `Timeouts` class,
            which allows setting the timeout for each operation independently.
        host_health_monitor: A `.BasicHealthMonitor` is used by default, so
            your tests can control whether the session sees the machine as
            healthy or not by calling `.set_healthy` and `.set_unhealthy` on
            that instance.  If you instead pass `None`, the session will always
            see the machine as healthy, `.HostUnhealthy` and
            `.HostHealthRestored` events will never be emitted, and the
            *suspends_on_bad_host_health* option of `QueueOptions` cannot be
            used.
        num_processing_threads: The number of threads for the SDK to use for
            processing events.  This defaults to 1.
        blob_buffer_size: The size (in bytes) of the blob buffers to use.  This
            defaults to 4k.
        channel_high_watermark: The size (in bytes) to use for the write cache
            high watermark on the channel.  The default value is 128MB.  Note
            that BlazingMQ reserves 4MB of this value for control messages, so
            the actual watermark for data published is
            ``channel_high_watermark - 4MB``.
        event_queue_watermarks: A tuple containing the low and high
            notification watermark thresholds for the buffer containing all
            incoming messages from the broker, respectively.  A warning
            `.SlowConsumerHighWaterMark` is emitted when the buffer reaches the
            high watermark value, and a notification `.SlowConsumerNormal` is
            emitted when the buffer is back to the low watermark.
        stats_dump_interval: The interval (in seconds) at which to dump stats
            into the logs.  If 0, disable the recurring dump of stats (final
            stats are always dumped at the end of the session).  The default is
            5min; the value must be a multiple of 30s, in the range
            ``[0s - 60min]``.

    Raises:
        `~blazingmq.Error`: If the session start request was not successful.
        `~blazingmq.exceptions.BrokerTimeoutError`: If the broker didn't respond
            to the request within a reasonable amount of time.
        `ValueError`: If any of the timeouts are provided and not > 0.0, or if
            the ``stats_dump_interval`` is provided and is < 0.0.
    """

    def __init__(
        self,
        on_session_event: Callable[[SessionEvent], None],
        on_message: Optional[Callable[[Message, MessageHandle], None]] = None,
        broker: str = "tcp://localhost:30114",
        message_compression_algorithm: CompressionAlgorithmType = (
            CompressionAlgorithmType.NONE
        ),
        timeout: Union[Timeouts, float] = DEFAULT_TIMEOUT,
        host_health_monitor: Union[BasicHealthMonitor, None] = (DefaultMonitor()),
        num_processing_threads: Optional[int] = None,
        blob_buffer_size: Optional[int] = None,
        channel_high_watermark: Optional[int] = None,
        event_queue_watermarks: Optional[tuple[int, int]] = None,
        stats_dump_interval: Optional[float] = None,
    ) -> None:
        if host_health_monitor is not None:
            if not isinstance(host_health_monitor, BasicHealthMonitor):
                raise TypeError(
                    f"host_health_monitor must be None or an instance of "
                    f"{' or '.join(KNOWN_MONITORS)}"
                )

        monitor_host_health = host_health_monitor is not None
        fake_host_health_monitor = getattr(host_health_monitor, "_monitor", None)

        self._has_no_on_message = on_message is None

        # Using our Timeouts class, preserve the old behavior of passing in a
        # simple float as a timeout.  Avoid setting the `connect_timeout` and
        # `disconnect_timeout`.
        if not isinstance(timeout, Timeouts):
            timeout = Timeouts(
                open_queue_timeout=timeout,
                configure_queue_timeout=timeout,
                close_queue_timeout=timeout,
            )

        self._ext = ExtSession(
            on_session_event,
            on_message=on_message,
            broker=six.ensure_binary(broker),
            message_compression_algorithm=message_compression_algorithm,
            num_processing_threads=num_processing_threads,
            blob_buffer_size=blob_buffer_size,
            channel_high_watermark=channel_high_watermark,
            event_queue_watermarks=event_queue_watermarks,
            stats_dump_interval=_convert_stats_dump_interval(stats_dump_interval),
            timeouts=_validate_timeouts(timeout),
            monitor_host_health=monitor_host_health,
            fake_host_health_monitor=fake_host_health_monitor,
        )

    @classmethod
    def with_options(
        cls,
        on_session_event: Callable[[SessionEvent], None],
        on_message: Optional[Callable[[Message, MessageHandle], None]] = None,
        broker: str = "tcp://localhost:30114",
        session_options: SessionOptions = (SessionOptions()),
    ) -> Session:
        """Construct a *Session* instance using `.SessionOptions`.

        This is the recommended way to construct a new session, as the
        `.SessionOptions` class provides an easier to use interface for
        configuring only those options you need.

        Args:
            on_session_event: a required callback to process `.SessionEvent` events
                received by the session.
            on_message: an optional callback to process `Message` objects received
                by the session.
            broker: TCP address of the broker (default: 'tcp://localhost:30114').
                If the environment variable ``BMQ_BROKER_URI`` is set, its value
                will override whatever broker address is passed via this argument.
            session_options: an instance of `.SessionOptions` that represents the
                session's configuration.

        Raises:
            `~blazingmq.Error`: If the session start request was not successful.
            `~blazingmq.exceptions.BrokerTimeoutError`: If the broker didn't respond
                to the request within a reasonable amount of time.
            `ValueError`: If any of the timeouts are provided and not > 0.0, or if
                the ``stats_dump_interval`` is provided and is < 0.0.
        """
        message_compression_algorithm = session_options.message_compression_algorithm
        if message_compression_algorithm is None:
            message_compression_algorithm = CompressionAlgorithmType.NONE

        if session_options.timeouts is None:
            return cls(
                on_session_event,
                on_message,
                broker,
                message_compression_algorithm,
                DEFAULT_TIMEOUT,
                session_options.host_health_monitor,
                session_options.num_processing_threads,
                session_options.blob_buffer_size,
                session_options.channel_high_watermark,
                session_options.event_queue_watermarks,
                session_options.stats_dump_interval,
            )
        else:
            return cls(
                on_session_event,
                on_message,
                broker,
                message_compression_algorithm,
                session_options.timeouts,
                session_options.host_health_monitor,
                session_options.num_processing_threads,
                session_options.blob_buffer_size,
                session_options.channel_high_watermark,
                session_options.event_queue_watermarks,
                session_options.stats_dump_interval,
            )

    def open_queue(
        self,
        queue_uri: str,
        read: bool = False,
        write: bool = False,
        options: QueueOptions = QueueOptions(),
        timeout: float = DEFAULT_TIMEOUT,
    ) -> None:
        """Open a queue with the specified parameters

        Open a queue at the specified *queue_uri*. The *queue_uri* is the
        identifier for any future interactions with this opened queue.

        Note:
            Invoking this method from the ``on_message`` or
            ``on_session_event`` of the `Session` or the ``on_ack`` callback of
            a posted message will cause a deadlock.

        Note:
            It's possible to override the default tier by adding an optional
            tier to the queue URI.  See `URI format`__.

        __ https://bloomberg.github.io/blazingmq/docs/apidocs/cpp_apidocs/group__bmqt__uri.html

        Args:
            queue_uri: unique resource identifier for the queue to be opened.
            read: open the queue for reading, enabling the `Session` to receive
                `Message` objects for this queue.
            write: open the queue for writing, allowing posting to this queue.
            options (~blazingmq.QueueOptions): options to configure the queue
                with
            timeout: maximum number of seconds to wait for this request.
                If not provided, the *timeout* provided to the `Session` when
                it was created it used.  If that was not provided either,
                a reasonable default is used.

        Raises:
            `~blazingmq.Error`: If the open queue request was not successful.
            `~blazingmq.exceptions.BrokerTimeoutError`: If the broker didn't
                respond to the request within a reasonable amount of time.
            `ValueError`: If *timeout* is not > 0.0.
        """
        if read and self._has_no_on_message:
            raise Error(
                "Can't open queue {} in read mode: no on_message "
                "callback was provided at Session construction".format(queue_uri)
            )

        if options.suspends_on_bad_host_health and not self._ext.monitor_host_health:
            raise Error(
                "Queues cannot use suspends_on_bad_host_health if host health"
                " monitoring was disabled when the Session was created"
            )

        self._ext.open_queue_sync(
            six.ensure_binary(queue_uri),
            read=read,
            write=write,
            consumer_priority=options.consumer_priority,
            max_unconfirmed_messages=options.max_unconfirmed_messages,
            max_unconfirmed_bytes=options.max_unconfirmed_bytes,
            suspends_on_bad_host_health=options.suspends_on_bad_host_health,
            timeout=_convert_timeout(timeout),
        )

    def close_queue(self, queue_uri: str, timeout: float = DEFAULT_TIMEOUT) -> None:
        """Close an opened queue at the specified *queue_uri*.

        Close the queue at the specified *queue_uri*. After this method
        successfully returns, the *queue_uri* will no longer correspond to valid
        queue to do any operations.

        Note:
            Invoking this method from the ``on_message`` or
            ``on_session_event`` of the `Session` or the ``on_ack`` callback of
            a posted message will cause a deadlock.

        Args:
            queue_uri: unique resource identifier for the queue to be closed.
            timeout: maximum number of seconds to wait for this request.
                If not provided, the *timeout* provided to the `Session` when
                it was created it used.  If that was not provided either,
                a reasonable default is used.

        Raises:
            `~blazingmq.Error`: If the close queue request was not successful.
            `~blazingmq.exceptions.BrokerTimeoutError`: If the broker didn't
                respond to the request within a reasonable amount of time.
            `ValueError`: If *timeout* is not > 0.0.
        """
        self._ext.close_queue_sync(
            six.ensure_binary(queue_uri),
            timeout=_convert_timeout(timeout),
        )

    def configure_queue(
        self,
        queue_uri: str,
        options: QueueOptions,
        timeout: float = DEFAULT_TIMEOUT,
    ) -> None:
        """Modify the options on an opened queue at the specified *queue_uri*.

        Modify the options of the queue at the specified *queue_uri*. After this
        method successfully returns, the *queue_uri* will be configured with the
        specified *options*.

        Note:
            Invoking this method from the ``on_message`` or ``on_session_event`` of
            the `Session` or the ``on_ack`` callback of a posted message will
            cause a deadlock.

        Args:
            queue_uri: unique resource identifier for the queue to be configured.
            options (~blazingmq.QueueOptions): options to configure the queue with
            timeout: maximum number of seconds to wait for this request.
                If not provided, the *timeout* provided to the `Session` when
                it was created it used.  If that was not provided either,
                a reasonable default is used.

        Raises:
            `~blazingmq.Error`: If the configure queue request was not successful.
            `~blazingmq.exceptions.BrokerTimeoutError`: If the broker didn't
                respond to the request within a reasonable amount of time.
            `ValueError`: If *timeout* is not > 0.0.
        """
        if options.suspends_on_bad_host_health and not self._ext.monitor_host_health:
            raise Error(
                "Queues cannot use suspends_on_bad_host_health if host health"
                " monitoring was disabled when the Session was created"
            )

        self._ext.configure_queue_sync(
            six.ensure_binary(queue_uri),
            consumer_priority=options.consumer_priority,
            max_unconfirmed_messages=options.max_unconfirmed_messages,
            max_unconfirmed_bytes=options.max_unconfirmed_bytes,
            suspends_on_bad_host_health=options.suspends_on_bad_host_health,
            timeout=_convert_timeout(timeout),
        )

    def get_queue_options(self, queue_uri: str) -> QueueOptions:
        """Get configured options of an opened queue.

        Get the previously set options of an opened queue at the specified
        *queue_uri*.

        Args:
            queue_uri: unique resource identifier for the queue to be configured.

        Returns:
            `~blazingmq.QueueOptions`: The queue's configured options.

        Raises:
            `~blazingmq.Error`: If the queue with the given URI is not open, or
                its options cannot be retrieved.

        Note:
            Options that only affect message consumption, including
            *consumer_priority*, *max_unconfirmed_messages*, and
            *max_unconfirmed_bytes*, are ignored when opening or configuring
            a write-only queue, so any attempt to set those options on
            a write-only queue won't be reflected in the `QueueOptions`
            returned by a later call to *get_queue_options*.
        """
        options = self._ext.get_queue_options(six.ensure_binary(queue_uri))
        return QueueOptions(*options)

    def stop(self) -> None:
        """Teardown the broker connection

        Stop the session with the connected BlazingMQ broker which implies
        tearing down the connection. This session cannot be used to execute any
        actions after this method returns.

        Note:
            Invoking this method from the ``on_message`` or
            ``on_session_event`` of the `Session` or the ``on_ack`` callback of
            a posted message will cause a deadlock.
        """

        self._ext.stop()

    def post(
        self,
        queue_uri: str,
        message: bytes,
        properties: Optional[PropertyValueDict] = None,
        property_type_overrides: Optional[PropertyTypeDict] = None,
        on_ack: Optional[Callable[[Ack], None]] = None,
    ) -> None:
        """Post a message to an opened queue specified by *queue_uri*.

        Post the payload and optional properties and overrides to the opened
        queue specified by *queue_uri*. Optionally take an *on_ack* callback
        that is invoked with the incoming acknowledgment for the message posted.

        Args:
            queue_uri: unique resource identifier for the queue to posted to.
            message: the payload of the message.
            properties (Optional[`~blazingmq.PropertyValueDict`]): optionally
                provided properties to be associated with the message.
            property_type_overrides (Optional[`~blazingmq.PropertyTypeDict`]):
                optionally provided type overrides for the properties.
            on_ack (Optional[Callable[[~blazingmq.Ack], None]]): optionally
                specified callback which is invoked with the acknowledgment
                status of the message being posted.

        Raises:
            `~blazingmq.Error`: If the post request was not successful.
        """
        props: Optional[Dict[bytes, Tuple[Union[int, bytes], int]]] = None
        if properties or property_type_overrides:
            props = _collect_properties_and_types(properties, property_type_overrides)

        self._ext.post(
            six.ensure_binary(queue_uri),
            message,
            properties=props,
            on_ack=on_ack,
        )

    def confirm(self, message: Message) -> None:
        """Confirm the specified message from this queue.

        Mark the specified *message*, as confirmed. If successful, this will
        indicate to the BlazingMQ framework that the processing for this
        message has been completed and that the message must not be
        re-delivered, if this queue is opened again by any consumer.

        It's often more convenient to use `MessageHandle.confirm`
        instead. An instance of `MessageHandle` is received with every
        new `Message` delivered via the ``on_message`` callback.

        Args:
            message (~blazingmq.Message): message to be confirmed.

        Raises:
            `~blazingmq.Error`: If the confirm message request was not
                successful.
        """
        self._ext.confirm(message)

    def __enter__(self) -> Session:
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, exc_traceback: Any) -> None:
        self.stop()



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

cimport bsl
cimport bsl.bslma as bslma


cdef extern from 'bdem_schema.h' nogil:

    cdef cppclass Schema 'BloombergLP::bdem_Schema':
        pass


cdef extern from 'bdem_berencoderoptions.h' nogil:

    cdef cppclass BerEncoderOptions 'BloombergLP::bdem_BerEncoderOptions':
        pass


cdef extern from 'bdem_berencoder.h' nogil:

    cdef cppclass BerEncoder 'BloombergLP::bdem_BerEncoder':

        # CREATORS
        BerEncoder(const BerEncoderOptions *options=0,
                   bslma.Allocator *allocator=0)

        int encode[TYPE](bsl.streambuf *streamBuf, const TYPE& value)

        bsl.StringRef loggedMessages() const


cdef extern from 'bdem_elemtype.h' nogil:

    cdef enum e_ElemType 'BloombergLP::bdem_ElemType::Type':
        # SCALAR
        BDEM_CHAR       'BloombergLP::bdem_ElemType::BDEM_CHAR'
        BDEM_SHORT      'BloombergLP::bdem_ElemType::BDEM_SHORT'
        BDEM_INT        'BloombergLP::bdem_ElemType::BDEM_INT'
        BDEM_INT64      'BloombergLP::bdem_ElemType::BDEM_INT64'
        BDEM_FLOAT      'BloombergLP::bdem_ElemType::BDEM_FLOAT'
        BDEM_DOUBLE     'BloombergLP::bdem_ElemType::BDEM_DOUBLE'
        BDEM_BOOL       'BloombergLP::bdem_ElemType::BDEM_BOOL'
        BDEM_STRING     'BloombergLP::bdem_ElemType::BDEM_STRING'
        BDEM_DATETIME   'BloombergLP::bdem_ElemType::BDEM_DATETIME'
        BDEM_DATETIMETZ 'BloombergLP::bdem_ElemType::BDEM_DATETIMETZ'
        BDEM_DATE       'BloombergLP::bdem_ElemType::BDEM_DATE'
        BDEM_DATETZ     'BloombergLP::bdem_ElemType::BDEM_DATETZ'
        BDEM_TIME       'BloombergLP::bdem_ElemType::BDEM_TIME'
        BDEM_TIMETZ     'BloombergLP::bdem_ElemType::BDEM_TIMETZ'

        # ARRAY
        BDEM_BOOL_ARRAY 'BloombergLP::bdem_ElemType::BDEM_BOOL_ARRAY'
        BDEM_CHAR_ARRAY 'BloombergLP::bdem_ElemType::BDEM_CHAR_ARRAY'



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from bsl cimport optional
from bsl cimport pair
from bsl cimport shared_ptr
from bsl.bsls cimport TimeInterval
from libcpp cimport bool as cppbool

from bmq.bmqa cimport ManualHostHealthMonitor
from bmq.bmqt cimport CompressionAlgorithmType


cdef extern from 'pybmq_ballutil.h' namespace 'BloombergLP::pybmq':
    cdef cppclass BallUtil:
        @staticmethod
        object initBallSingleton(
            object (*cb)(char*, int, char*, int, char*),
            object context,
        ) except +

        @staticmethod
        object shutDownBallSingleton() except +

cdef extern from "pybmq_session.h" namespace "BloombergLP::pybmq" nogil:
    cdef cppclass Session:
        Session(object on_session_event,
                object on_message_event,
                object on_ack_event,
                const char* broker_uri,
                const char* script_name,
                CompressionAlgorithmType message_compression_algorithm,
                optional[int] num_processing_threads,
                optional[int] blob_buffer_size,
                optional[int] channel_high_watermark,
                optional[pair[int, int]] event_queue_watermarks,
                TimeInterval stats_dump_interval,
                TimeInterval connect_timeout,
                TimeInterval disconnect_timeout,
                TimeInterval open_queue_timeout,
                TimeInterval configure_queue_timeout,
                TimeInterval close_queue_timeout,
                bint monitor_host_health,
                shared_ptr[ManualHostHealthMonitor] fake_host_health_monitor_sp,
                object error,
                object broker_timeout_error,
                object mock) except+

        object start(TimeInterval) except+
        object stop(bint) except+

        object open_queue_sync(const char* queue_uri,
                               bint read,
                               bint write,
                               optional[int] consumer_priority,
                               optional[int] max_unconfirmed_messages,
                               optional[int] max_unconfirmed_bytes,
                               optional[cppbool] suspends_on_bad_host_health,
                               TimeInterval timeout) except+

        object configure_queue_sync(const char* queue_uri,
                                    optional[int] consumer_priority,
                                    optional[int] max_unconfirmed_messages,
                                    optional[int] max_unconfirmed_bytes,
                                    optional[cppbool] suspends_on_bad_host_health,
                                    TimeInterval timeout) except+

        object close_queue_sync(const char* queue_uri, TimeInterval timeout) except+

        object get_queue_options(const char* queue_uri) except+

        object post(const char* queue_uri,
                    const char* payload,
                    size_t payload_length,
                    object properties,
                    object on_ack) except+
        object confirm(const char* queue_uri, const unsigned char* guid, size_t guid_length) except+



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class AckStatus(Enum):
    """An enum representing the status of an Ack message

    An `AckStatus` is a status of a received `Ack` message
    which is the result of an attempted put to some particular queue.
    Anything other than `AckStatus.SUCCESS` represents a failure.
    """

    SUCCESS = object()
    UNKNOWN = object()
    TIMEOUT = object()
    NOT_CONNECTED = object()
    CANCELED = object()
    NOT_SUPPORTED = object()
    REFUSED = object()
    INVALID_ARGUMENT = object()
    NOT_READY = object()
    LIMIT_BYTES = object()
    LIMIT_MESSAGES = object()
    STORAGE_FAILURE = object()
    UNRECOGNIZED = object()
    """The `AckStatus` was not recognized by the binding layer"""

    def __repr__(self) -> str:
        # hide the unimportant value of `object()`
        return f"<{self.__class__.__name__}.{self.name}>"


class CompressionAlgorithmType(Enum):
    """An enum representing compression algorithm used by a producer"""

    NONE = object()
    ZLIB = object()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}.{self.name}>"


class PropertyType(Enum):
    """An enum representing various data types understood by BlazingMQ"""

    BOOL = object()
    CHAR = object()
    SHORT = object()
    INT32 = object()
    INT64 = object()
    STRING = object()
    BINARY = object()

    def __repr__(self) -> str:
        # hide the unimportant value of `object()`
        return f"<{self.__class__.__name__}.{self.name}>"



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
from typing import Optional

LOGGER = logging.getLogger(__name__)


class SessionEvent:
    """Base session event type"""

    def __init__(self, message: Optional[str]) -> None:
        self._message = message

    def __repr__(self) -> str:
        if self._message:
            return f"<{self.__class__.__name__}: {self._message}>"
        else:
            return f"<{self.__class__.__name__}>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SessionEvent):
            return False
        return self.__class__ is other.__class__ and self._message == other._message

    def __ne__(self, other: object) -> bool:
        return not self == other


class QueueEvent(SessionEvent):
    """Base type for session events relating to a single queue.

    Attributes:
        queue_uri (str): Queue URI this event is associated with.
    """

    def __init__(self, queue_uri: str, message: Optional[str] = None) -> None:
        self.queue_uri = queue_uri
        super().__init__(message)

    def __repr__(self) -> str:
        if self._message:
            return "<{}: {} {}>".format(
                self.__class__.__name__, self.queue_uri, self._message
            )
        else:
            return f"<{self.__class__.__name__}: {self.queue_uri}>"

    def __eq__(self, other: object) -> bool:
        if type(self) is not type(other):
            return NotImplemented

        assert isinstance(other, QueueEvent)  # for mypy's sake
        return (
            self.__class__ is other.__class__
            and self._message == other._message
            and self.queue_uri == other.queue_uri
        )


class Connected(SessionEvent):
    """Notification of successful connection with the broker"""


class Disconnected(SessionEvent):
    """Notification of successful disconnection with the broker"""


class ConnectionLost(SessionEvent):
    """Notification of a lost connection with the broker"""


class Reconnected(SessionEvent):
    """Notification of a re-connection with the broker in case connection was lost earlier"""


class StateRestored(SessionEvent):
    """Notification of successfully restoring state of application as it was before lost
    connection or disconnection"""


class ConnectionTimeout(SessionEvent):
    """Notification that a requested connection could not be initiated in within the
    timeout period"""


class HostUnhealthy(SessionEvent):
    """Notification that the host has been marked unhealthy.

    This is emitted only if ``host_health_monitor=None`` was not provided when
    the `.Session` was created.  This will be emitted whenever the machine
    becomes unhealthy.  It is also emitted if the machine is initially unhealthy
    when the `.Session` is created.

    .. versionadded:: 0.7.0
    """


class HostHealthRestored(SessionEvent):
    """Notification that the host is no longer marked unhealthy.

    This is emitted only if ``host_health_monitor=None`` was not provided when
    the `.Session` was created.  It will be emitted once the machine is becomes
    healthy after an earlier `.HostUnhealthy` event. Before this event
    is emitted, you will receive a `QueueResumed` or `QueueResumeFailed` for
    each queue that was suspended due to ``suspends_on_bad_host_health=True``.

    .. versionadded:: 0.7.0
    """


class QueueSuspended(QueueEvent):
    """A queue that is sensitive to host health has been suspended.

    After a `.HostUnhealthy` event is emitted, any queue that was opened with
    ``suspend_on_bad_host_health=True`` will suspend operation. This event will
    be emitted once for each suspended queue.

    Note:
        If ``host_health_monitor=None`` was provided when the `.Session` was
        created, this event will never be emitted because the host will never
        be considered unhealthy.

    Attributes:
        queue_uri (str): URI of the queue that has been successfully suspended.

    .. versionadded:: 0.7.0
    """


class QueueSuspendFailed(QueueEvent):
    """A queue that is sensitive to host health could not be suspended.

    Whenever a `QueueSuspended` event would be expected, this event may be
    emitted instead if the SDK is unable to suspend the queue as expected.

    Note:
        The BlazingMQ SDK considers the failure to suspend a queue as evidence
        of an unusually serious problem with the connection to the broker, so
        if this event occurs the SDK follows it up by dropping the connection
        to the broker and trying to re-establish it.

    Attributes:
        queue_uri (str): URI of the queue that could not be suspended.

    .. versionadded:: 0.7.0
    """


class QueueResumed(QueueEvent):
    """A queue that is sensitive to host health has been resumed.

    Once an unhealthy machine becomes healthy again, the SDK will automatically
    attempt to resume each queue that was suspended when the machine became
    unhealthy. This event will be emitted once for each queue that had been
    suspended, only after which will `HostHealthRestored` be emitted.

    Attributes:
        queue_uri (str): URI of the queue that has been successfully resumed.

    .. versionadded:: 0.7.0
    """


class QueueResumeFailed(QueueEvent):
    """A queue that is sensitive to host health could not be resumed.

    Whenever a `QueueResumed` event would be expected, this event may be
    emitted instead if the SDK is unable to resume the queue as expected.

    Note:
        Unlike if suspending a queue fails, the SDK will not automatically drop
        the connection to the broker if resuming a queue fails.

    Attributes:
        queue_uri (str): URI of the queue that could not be resumed.

    .. versionadded:: 0.7.0
    """


class QueueReopened(QueueEvent):
    """A queue has been successfully reopened after a connection loss.

    If the connection with the broker is lost, `ConnectionLost` is emitted.
    Once it is reestablished, `Reconnected` is emitted, followed by either
    a `QueueReopened` or `QueueReopenFailed` for each queue that was
    previously open, and finally `StateRestored` is emitted.

    Attributes:
        queue_uri (str): URI of the queue that has been successfully reopened.
    """


class QueueReopenFailed(QueueEvent):
    """A queue couldn't be reopened after a connection loss.

    Attributes:
        queue_uri (str): URI of the queue that could not be reopened.
    """


class SlowConsumerNormal(SessionEvent):
    """Notification that the consumer has resumed acceptable rate of consumption"""


class SlowConsumerHighWaterMark(SessionEvent):
    """Notification that the consumer is consuming at the lowest rate acceptable"""


class Error(SessionEvent):
    """Notification of a miscellaneous error"""


class InterfaceError(SessionEvent):
    """The BlazingMQ SDK behaved in an unexpected way."""


def log_session_event(event: SessionEvent) -> None:
    """Log incoming session event as appropriate

    A callback that can be used as a default for the *on_session_event*
    parameter on the `.Session` object. All `.Connected`, `.Disconnected`,
    `.StateRestored`, `.SlowConsumerNormal`, and `.QueueReopened` events are
    logged at INFO level, and any `.ConnectionLost`, `.SlowConsumerHighWaterMark`,
    and `.Reconnected` events are logged at WARN level, as they may indicate issues
    with the application. Any other events are most likely an error in the
    application and are logged at ERROR level.

    Args:
        event (~blazingmq.session_events.SessionEvent): incoming `SessionEvent`
            object.
    """
    if isinstance(
        event,
        (
            Connected,
            Disconnected,
            StateRestored,
            SlowConsumerNormal,
            QueueReopened,
            HostUnhealthy,
            HostHealthRestored,
            QueueSuspended,
            QueueResumed,
        ),
    ):
        level = logging.INFO
    elif isinstance(event, (ConnectionLost, Reconnected, SlowConsumerHighWaterMark)):
        level = logging.WARN
    else:
        # ConnectionTimeout, Error, InterfaceError, QueueReopenFailed,
        # QueueSuspendFailed, QueueResumeFailed
        level = logging.ERROR

    LOGGER.log(level, "Received session event: %s", event)



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ._ext import FakeHostHealthMonitor


class BasicHealthMonitor:
    """Control whether a `.Session` sees the host as healthy or unhealthy.

    When a *BasicHealthMonitor* is passed for the `.Session` constructor's
    *host_health_monitor* parameter, you can control whether the BlazingMQ
    session sees the host as healthy or unhealthy by calling the `.set_healthy`
    and `.set_unhealthy` methods.  Newly created instances default to the
    healthy state.
    """

    def __init__(self) -> None:
        self._monitor = FakeHostHealthMonitor()

    def set_healthy(self) -> None:
        """Tell any associated `.Session` that the host is healthy."""
        self._monitor.set_healthy()

    def set_unhealthy(self) -> None:
        """Tell any associated `.Session` that the host is unhealthy."""
        self._monitor.set_unhealthy()

    def __repr__(self) -> str:
        return "BasicHealthMonitor()"



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Optional
from typing import TYPE_CHECKING

from ._enums import AckStatus
from ._typing import PropertyTypeDict
from ._typing import PropertyValueDict
from .exceptions import Error

if TYPE_CHECKING:
    # Safely perform circular references only during static type analysis
    from . import _ext  # pragma: no cover


def pretty_hex(blob: bytes) -> str:
    return blob.hex().upper()


def create_message(
    data: bytes,
    guid: bytes,
    queue_uri: str,
    properties: PropertyValueDict,
    property_types: PropertyTypeDict,
) -> Message:
    inst = Message.__new__(Message)
    assert isinstance(inst, Message)
    inst._set_attrs(data, guid, queue_uri, properties, property_types)
    return inst


class Message:
    """A class representing a message received from BlazingMQ.

    A `Message` represents a message delivered by BlazingMQ from a producer
    to this queue. This message can only be received if the queue is
    opened with 'read=True' mode enabled.

    Attributes:
        data (bytes): Payload for the message received from BlazingMQ.
        guid (bytes): Globally unique id for this message.
        queue_uri (str): Queue URI this message is for.
        properties (dict): A dictionary of BlazingMQ message properties.
            The dictionary keys must be :class:`str` representing the property
            names and the values must be of type :class:`str`, :class:`bytes`,
            :class:`bool` or :class:`int`.
        property_types (dict): A mapping of property names to
            `PropertyType` types. The dictionary is guaranteed to provide
            a value for each key already present in `Message.properties`
    """

    def _set_attrs(
        self,
        data: bytes,
        guid: bytes,
        queue_uri: str,
        properties: PropertyValueDict,
        property_types: PropertyTypeDict,
    ) -> None:
        """Teach mypy what our instance variables are despite our private __init__"""
        self.data = data
        self.guid = guid
        self.queue_uri = queue_uri
        self.properties = properties
        self.property_types = property_types

    def __init__(self) -> None:
        raise Error("The Message class does not have a public constructor.")

    def __repr__(self) -> str:
        return f"<Message[{pretty_hex(self.guid)}] for {self.queue_uri}>"


def create_message_handle(message: Message, ext_session: _ext.Session) -> MessageHandle:
    inst = MessageHandle.__new__(MessageHandle)
    assert isinstance(inst, MessageHandle)
    inst._set_attrs(message, ext_session)
    return inst


class MessageHandle:
    """Operations that can be performed on a `Message`.

    An instance of this class is received in the ``on_message``
    callback along with an instance of a `Message`.
    """

    def confirm(self) -> None:
        """Confirm the message received along with this handle.

        See `Session.confirm` for more details.

        Raises:
            `~blazingmq.Error`: If the confirm message request
                was not successful.
        """
        self._ext_session.confirm(self._message)

    def _set_attrs(self, message: Message, ext_session: _ext.Session) -> None:
        """Teach mypy what our instance variables are despite our private __init__"""
        self._message = message
        self._ext_session = ext_session

    def __init__(self) -> None:
        raise Error("The MessageHandle class does not have a public constructor.")

    def __repr__(self) -> str:
        return "<MessageHandle[{}] for {}>".format(
            pretty_hex(self._message.guid),
            self._message.queue_uri,
        )


def create_ack(
    guid: Optional[bytes], status: AckStatus, status_description: str, queue_uri: str
) -> Ack:
    inst = Ack.__new__(Ack)
    assert isinstance(inst, Ack)
    inst._set_attrs(guid, status, status_description, queue_uri)
    return inst


class Ack:
    """Acknowledgment message

    An `Ack` is a notification from BlazingMQ to the application,
    specifying that the message has been received. This is valuable
    for ensuring delivery of messages.

    These messages will be received in the optionally provided callback to
    `Session.post()`.

    An `Ack` is by itself not an indication of success unless it has a status of
    `AckStatus.SUCCESS`.

    Attributes:
        guid (bytes): a globally unique identifier generated by BlazingMQ for the
            message that was successfully posted. This can be correlated between the
            producer and consumer to verify the flow of messages.
        queue_uri (str): the queue that this message was routed to. This is useful
            if you have many queues and you want to route this particular `Ack` to a
            particular queue.
        status (AckStatus): the `AckStatus` indicating the result of the post
            operation. Unless this is of type `AckStatus.SUCCESS`, the post has
            failed and potentially needs to be dealt with.
    """

    def _set_attrs(
        self,
        guid: Optional[bytes],
        status: AckStatus,
        status_description: str,
        queue_uri: str,
    ) -> None:
        """Teach mypy what our instance variables are despite our private __init__"""
        self.guid = guid
        self.status = status
        self._status_description = status_description
        self.queue_uri = queue_uri

    def __init__(self) -> None:
        raise Error("The Ack class does not have a public constructor.")

    def __repr__(self) -> str:
        guid_identifier = "" if self.guid is None else f"[{pretty_hex(self.guid)}]"
        return "<Ack{} {} for {}>".format(
            guid_identifier,
            self._status_description,
            self.queue_uri,
        )



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcpp cimport bool


cdef extern from "bmqt_sessioneventtype.h" namespace "BloombergLP::bmqt::SessionEventType" nogil:
    cdef enum SessionEventEnum "BloombergLP::bmqt::SessionEventType::Enum":
        e_CONNECTED
        e_DISCONNECTED
        e_CONNECTION_LOST
        e_RECONNECTED
        e_STATE_RESTORED
        e_HOST_UNHEALTHY
        e_HOST_HEALTH_RESTORED
        e_QUEUE_SUSPENDED
        e_QUEUE_RESUMED
        e_CONNECTION_TIMEOUT
        e_QUEUE_REOPEN_RESULT
        e_SLOWCONSUMER_NORMAL
        e_SLOWCONSUMER_HIGHWATERMARK
        e_ERROR

cdef extern from "bmqt_resultcode.h" namespace "BloombergLP::bmqt::AckResult" nogil:
    cdef enum AckResult "BloombergLP::bmqt::AckResult::Enum":
        e_SUCCESS
        e_UNKNOWN
        e_TIMEOUT
        e_NOT_CONNECTED
        e_CANCELED
        e_NOT_SUPPORTED
        e_REFUSED
        e_INVALID_ARGUMENT
        e_NOT_READY
        e_LIMIT_MESSAGES
        e_LIMIT_BYTES
        e_STORAGE_FAILURE

cdef extern from "bmqt_propertytype.h" namespace "BloombergLP::bmqt::PropertyType" nogil:
    cdef enum PropertyType "BloombergLP::bmqt::PropertyType::Enum":
        e_BOOL
        e_CHAR
        e_SHORT
        e_INT32
        e_INT64
        e_STRING
        e_BINARY

cdef extern from "bmqt_compressionalgorithmtype.h" namespace "BloombergLP::bmqt::CompressionAlgorithmType" nogil:  # noqa: E501
    cdef enum CompressionAlgorithmType "BloombergLP::bmqt::CompressionAlgorithmType::Enum":
        e_NONE
        e_ZLIB

cdef extern from "bmqt_hosthealthstate.h" namespace "BloombergLP::bmqt::HostHealthState" nogil:  # noqa: E501
    cdef enum HostHealthState "BloombergLP::bmqt::HostHealthState::Enum":
        e_HEALTHY
        e_UNHEALTHY

cdef extern from "bmqt_queueoptions.h" namespace "BloombergLP::bmqt::QueueOptions" nogil:  # noqa: E501
    int k_DEFAULT_MAX_UNCONFIRMED_MESSAGES
    int k_DEFAULT_MAX_UNCONFIRMED_BYTES
    int k_DEFAULT_CONSUMER_PRIORITY
    bool k_DEFAULT_SUSPENDS_ON_BAD_HOST_HEALTH



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
import platform
import sys

import pkgconfig
from setuptools import Extension
from setuptools import setup

# XXX: Cython imports must come after importing setuptools
from Cython.Build import cythonize  # isort:skip
from Cython.Compiler import Options  # isort:skip

IS_MAC = platform.system() == "Darwin"
IS_LINUX = platform.system() == "Linux"

# Making an sdist in principle does not require BDE+NTF+BlazingMQ. However,
# every time setup.py is run, it calls create_extension, which searches for bmq
# with pkgconfig. This is generally good; if we're building the SDK of course
# we need BlazingMQ built and installed on the system. For building an sdist,
# though, this is not needed.
#
# On GitHub CI, the sdist is generated in a separate container, before we run
# build_manylinux.sh to build the wheels in another container. We could run
# build_manylinux.sh prior to making the sdist, but that's a 30-minute
# waste. Instead, with this environment variable, we can selectively tell
# setup.py to not look for pkgconfig dependencies. This is only safe when
# making sdists, but that is exactly where we want it to happen.
SHOULD_SKIP_PKGCONFIG = "BLAZINGMQ_PYTHON_NO_PKGCONFIG" in os.environ

TEST_BUILD = False
if "--test-build" in sys.argv:
    TEST_BUILD = True
    sys.argv.remove("--test-build")

if os.getenv("CYTHON_TEST_MACROS", None) is not None:
    TEST_BUILD = True


COMPILER_DIRECTIVES = {
    "language_level": "3str",
    "embedsignature": True,
    "boundscheck": False,
    "wraparound": False,
    "cdivision": True,
    "linetrace": False,
    "c_string_type": "unicode",
    "c_string_encoding": "utf8",
}

DEFINE_MACROS = []

if TEST_BUILD:
    COMPILER_DIRECTIVES = {
        "language_level": "3str",
        "boundscheck": True,
        "embedsignature": True,
        "wraparound": True,
        "cdivision": False,
        "linetrace": True,
        "overflowcheck": True,
        "infer_types": True,
        "c_string_type": "unicode",
        "c_string_encoding": "utf8",
    }
    DEFINE_MACROS.extend([("CYTHON_TRACE", "1"), ("CYTHON_TRACE_NOGIL", "1")])


def create_extension(name, libraries, **kwargs):
    extra_compile_args = []

    # Enable extra warnings.
    extra_compile_args.extend(
        [
            "-Wall",
            "-Wextra",
            "-Wno-invalid-offsetof",
            "-Wno-missing-field-initializers",
            "-Wno-deprecated-declarations",
        ]
    )

    # Work around unused warning when built with old versions of Cython.
    # See https://github.com/cython/cython/issues/4948 for details.
    if sys.version_info[0:2] >= (3, 11):
        extra_compile_args.append("-DCYTHON_NCP_UNUSED=CYTHON_UNUSED")

    # Hardwire C++ flags needed for BDE bslstl support.
    extra_compile_args.extend(["-std=gnu++17", "-D_GLIBCXX_USE_CXX11_ABI=0"])

    extra_link_args = []

    if IS_LINUX:
        # Error if any of the shared libraries we link in has unresolved
        # symbols not satisfied by another library on the link line.
        extra_link_args.append("-Wl,--no-allow-shlib-undefined")

        # Hide symbols except for our required entry point.
        modname = name.rpartition(".")[-1]
        extra_link_args.append(f"-Wl,--version-script=src/blazingmq/{modname}.vers")
    elif IS_MAC:
        # Hide symbols except for our required entry point.
        modname = name.rpartition(".")[-1]
        extra_link_args.append(
            f"-Wl,-exported_symbols_list,src/blazingmq/{modname}.exp"
        )

    ext = Extension(name, **kwargs)
    ext.extra_compile_args = extra_compile_args + list(ext.extra_compile_args)
    ext.extra_link_args = extra_link_args + list(ext.extra_link_args)

    if SHOULD_SKIP_PKGCONFIG:
        pkg_config_libraries = ""
    else:
        pkg_config_libraries = " ".join(libraries)
    pkg_config = pkgconfig.parse(pkg_config_libraries, static=True)
    ext.define_macros = list(ext.define_macros) + pkg_config["define_macros"]
    ext.include_dirs = list(ext.include_dirs) + pkg_config["include_dirs"]
    ext.library_dirs = list(ext.library_dirs) + pkg_config["library_dirs"]

    dynamic_libs = {
        "crypt",
        "dl",
        "m",
        "pthread",
        "resolv",
        "rt",
        "util",
        "z",
    }
    ext.extra_link_args += [
        f"-l{lib}" if (IS_MAC or lib in dynamic_libs) else f"-l:lib{lib}.a"
        for lib in pkg_config["libraries"]
    ]

    return ext


EXTENSIONS = [
    create_extension(
        "blazingmq._ext",
        sources=[
            "src/blazingmq/_ext.pyx",
            "src/cpp/pybmq_ballutil.cpp",
            "src/cpp/pybmq_gilacquireguard.cpp",
            "src/cpp/pybmq_gilreleaseguard.cpp",
            "src/cpp/pybmq_messageutils.cpp",
            "src/cpp/pybmq_mocksession.cpp",
            "src/cpp/pybmq_refutils.cpp",
            "src/cpp/pybmq_session.cpp",
            "src/cpp/pybmq_sessioneventhandler.cpp",
        ],
        language="c++",
        include_dirs=["src/cpp", "src"],
        libraries=["bmq"],
        define_macros=DEFINE_MACROS,
    ),
]


Options.generate_cleanup_code = 3

if not IS_LINUX and not IS_MAC:
    raise RuntimeError(f"BlazingMQ does not support this platform ({sys.platform})")

about = {}
with open("src/blazingmq/_about.py") as fp:
    exec(fp.read(), about)

# Read the contents of the README file.
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="blazingmq",
    description="Python BlazingMQ API",
    version=about["__version__"],
    author="Bloomberg Finance LP",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed",
    ],
    package_data={"blazingmq": ["py.typed", "_ext.pyi"]},
    package_dir={"": "src"},
    packages=["blazingmq"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    ext_modules=cythonize(
        EXTENSIONS,
        include_path=["src/declarations"],
        compiler_directives=COMPILER_DIRECTIVES,
    ),
    python_requires=">=3.9",
    zip_safe=False,
)



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

cimport bsl.bslma as bslma


cdef extern from 'bsl_memory.h' namespace 'bsl' nogil:

    cdef cppclass shared_ptr[ELEMENT_TYPE]:
        # CREATORS
        shared_ptr()

        shared_ptr[ELEMENT_TYPE] shared_ptr[COMPATIBLE_TYPE, DELETER](
            COMPATIBLE_TYPE* ptr, DELETER *deleter)

        shared_ptr[ELEMENT_TYPE] shared_ptr[COMPATIBLE_TYPE](COMPATIBLE_TYPE* ptr)
        shared_ptr[ELEMENT_TYPE] shared_ptr[COMPATIBLE_TYPE](
            const shared_ptr[COMPATIBLE_TYPE]& other)

        # MANIPULATORS
        void createInplace(bslma.Allocator *allocator=0)

        # ACCESSORS
        ELEMENT_TYPE& operator*()

        ELEMENT_TYPE* get() const

        void reset()

cdef extern from 'bsl_memory.h' namespace 'bsl' nogil:

    cdef cppclass weak_ptr[ELEMENT_TYPE]:
        # CREATORS
        weak_ptr()

        weak_ptr[ELEMENT_TYPE] weak_ptr[COMPATIBLE_TYPE](
            const shared_ptr[COMPATIBLE_TYPE]& other)

        # ACCESSORS
        shared_ptr[ELEMENT_TYPE] lock() const

        bint expired() const

cdef extern from 'bsl_string.h' namespace 'bsl' nogil:

    cdef cppclass string:
        # CREATORS
        string(char *s)
        string()

        # MANIPULATORS
        string operator=(char *s)

        # ACCESSORS
        const char* c_str()

    cdef cppclass StringRef "BloombergLP::bslstl::StringRef":
        # CREATORS
        StringRef()

        StringRef(char *data)

        StringRef(char *data, int length)

        # StringRef(string& str)

        # ACCESSORS
        const char* data() const

        size_t length() const


cdef extern from 'bsl_streambuf.h' namespace 'bsl' nogil:

    cdef cppclass streambuf:
        pass


cdef extern from 'bsl_ostream.h' namespace 'bsl' nogil:

    cdef cppclass ostream:
        # FREE OPERATORS
        # Note that this is a declaration hack to workaround overloading
        # '::operator<<' per concrete type that would require circular
        # `cimport` statements.
        ostream& operator<<[TYPE](TYPE)

        string rdbuf() const


cdef extern from 'bsl_sstream.h' namespace 'bsl' nogil:

    cdef cppclass ostringstream(ostream):
        # ACCESSORS
        string str() const

    cdef cppclass stringstream:
        # CREATORS
        stringstream()

        # MANIPULATORS
        void str(const string buffer)

        # ACCESSORS
        streambuf* rdbuf()


cdef extern from 'bsl_vector.h' namespace 'bsl' nogil:

    cdef cppclass vector[T, ALLOCATOR=*]:
        ctypedef T value_type
        ctypedef ALLOCATOR allocator_type

        # these should really be allocator_type.size_type and
        # allocator_type.difference_type to be true to the C++ definition
        # but cython doesn't support deferred access on template arguments
        ctypedef size_t size_type
        ctypedef ptrdiff_t difference_type

        cppclass iterator:
            T& operator*()
            iterator operator++()
            iterator operator--()
            iterator operator+(size_type)
            iterator operator-(size_type)
            difference_type operator-(iterator)
            bint operator==(iterator)
            bint operator!=(iterator)
            bint operator<(iterator)
            bint operator>(iterator)
            bint operator<=(iterator)
            bint operator>=(iterator)
        cppclass reverse_iterator:
            T& operator*()
            reverse_iterator operator++()
            reverse_iterator operator--()
            reverse_iterator operator+(size_type)
            reverse_iterator operator-(size_type)
            difference_type operator-(reverse_iterator)
            bint operator==(reverse_iterator)
            bint operator!=(reverse_iterator)
            bint operator<(reverse_iterator)
            bint operator>(reverse_iterator)
            bint operator<=(reverse_iterator)
            bint operator>=(reverse_iterator)
        cppclass const_iterator(iterator):
            pass
        cppclass const_reverse_iterator(reverse_iterator):
            pass
        vector() except +
        vector(vector&) except +
        vector(size_type) except +
        vector(size_type, T&) except +
        vector(T*, T*) except +
        T& operator[](size_type)
        bint operator==(vector&, vector&)
        bint operator!=(vector&, vector&)
        bint operator<(vector&, vector&)
        bint operator>(vector&, vector&)
        bint operator<=(vector&, vector&)
        bint operator>=(vector&, vector&)
        void assign(size_type, const T&)
        void assign[input_iterator](input_iterator, input_iterator) except +
        T& at(size_type) except +
        T& back()
        iterator begin()
        const_iterator const_begin "begin"()
        size_type capacity()
        void clear()
        bint empty()
        iterator end()
        const_iterator const_end "end"()
        iterator erase(iterator)
        iterator erase(iterator, iterator)
        T& front()
        iterator insert(iterator, const T&) except +
        iterator insert(iterator, size_type, const T&) except +
        iterator insert[Iter](iterator, Iter, Iter) except +
        size_type max_size()
        void pop_back()
        void push_back(T&) except +
        reverse_iterator rbegin()
        const_reverse_iterator const_rbegin "crbegin"()
        reverse_iterator rend()
        const_reverse_iterator const_rend "crend"()
        void reserve(size_type)
        void resize(size_type) except +
        void resize(size_type, T&) except +
        size_type size()
        void swap(vector&)


cdef extern from "bsl_utility.h" namespace "bsl" nogil:
    cdef cppclass pair[T, U]:
        ctypedef T first_type
        ctypedef U second_type
        T first
        U second
        pair() except +
        pair(pair&) except +
        pair(T&, U&) except +
        bint operator==(pair&, pair&)
        bint operator!=(pair&, pair&)
        bint operator<(pair&, pair&)
        bint operator>(pair&, pair&)
        bint operator<=(pair&, pair&)
        bint operator>=(pair&, pair&)

cdef extern from 'bsl_set.h' namespace 'bsl' nogil:

    cdef cppclass set[T]:
        ctypedef T value_type
        cppclass iterator:
            T& operator*()
            iterator operator++()
            iterator operator--()
            bint operator==(iterator)
            bint operator!=(iterator)
        cppclass reverse_iterator:
            T& operator*()
            iterator operator++()
            iterator operator--()
            bint operator==(reverse_iterator)
            bint operator!=(reverse_iterator)
        cppclass const_iterator(iterator):
            pass
        cppclass const_reverse_iterator(reverse_iterator):
            pass
        set() except +
        set(set&) except +
        # set(key_compare&)
        # set& operator=(set&)
        bint operator==(set&, set&)
        bint operator!=(set&, set&)
        bint operator<(set&, set&)
        bint operator>(set&, set&)
        bint operator<=(set&, set&)
        bint operator>=(set&, set&)
        iterator begin()
        const_iterator const_begin "begin"()
        void clear()
        size_t count(const T&)
        bint empty()
        iterator end()
        const_iterator const_end "end"()
        pair[iterator, iterator] equal_range(const T&)
        # #pair[const_iterator, const_iterator] equal_range(T&)
        iterator erase(iterator)
        iterator erase(iterator, iterator)
        size_t erase(T&)
        iterator find(T&)
        const_iterator const_find "find"(T&)
        pair[iterator, bint] insert(const T&) except +
        iterator insert(iterator, const T&) except +
        void insert(iterator, iterator) except +
        # key_compare key_comp()
        iterator lower_bound(T&)
        const_iterator const_lower_bound "lower_bound"(T&)
        size_t max_size()
        reverse_iterator rbegin()
        const_reverse_iterator const_rbegin "rbegin"()
        reverse_iterator rend()
        const_reverse_iterator const_rend "rend"()
        size_t size()
        void swap(set&)
        iterator upper_bound(const T&)
        const_iterator const_upper_bound "upper_bound"(const T&)
        # value_compare value_comp()

cdef extern from "bsl_map.h" namespace "bsl" nogil:
    cdef cppclass map[T, U, COMPARE=*, ALLOCATOR=*]:
        ctypedef T key_type
        ctypedef U mapped_type
        ctypedef pair[const T, U] value_type
        ctypedef COMPARE key_compare
        ctypedef ALLOCATOR allocator_type
        cppclass iterator:
            pair[T, U]& operator*()
            iterator operator++()
            iterator operator--()
            bint operator==(iterator)
            bint operator!=(iterator)
        cppclass const_iterator:
            pair[const T, U]& operator*()
            const_iterator operator++()
            const_iterator operator--()
            bint operator==(const_iterator)
            bint operator!=(const_iterator)
        cppclass reverse_iterator:
            pair[T, U]& operator*()
            iterator operator++()
            iterator operator--()
            bint operator==(reverse_iterator)
            bint operator!=(reverse_iterator)
        cppclass const_reverse_iterator(reverse_iterator):
            pass
        map() except +
        map(map&) except +
        U& operator[](T&)
        bint operator==(map&, map&)
        bint operator!=(map&, map&)
        bint operator<(map&, map&)
        bint operator>(map&, map&)
        bint operator<=(map&, map&)
        bint operator>=(map&, map&)
        U& at(const T&) except +
        iterator begin()
        const_iterator const_begin "begin" ()
        void clear()
        size_t count(const T&)
        bint empty()
        iterator end()
        const_iterator const_end "end" ()
        pair[iterator, iterator] equal_range(const T&)
        void erase(iterator)
        void erase(iterator, iterator)
        size_t erase(const T&)
        iterator find(const T&)
        const_iterator const_find "find" (const T&)
        pair[iterator, bint] insert(pair[T, U]) except +  # XXX pair[T,U]&
        iterator insert(iterator, pair[T, U]) except +  # XXX pair[T,U]&
        iterator lower_bound(const T&)
        const_iterator const_lower_bound "lower_bound"(const T&)
        size_t max_size()
        reverse_iterator rbegin()
        const_reverse_iterator const_rbegin "rbegin"()
        reverse_iterator rend()
        const_reverse_iterator const_rend "rend"()
        size_t size()
        void swap(map&)
        iterator upper_bound(const T&)
        const_iterator const_upper_bound "upper_bound"(const T&)


cdef extern from "bsl_unordered_map.h" namespace "bsl" nogil:
    cdef cppclass unordered_map[T, U]:
        ctypedef T key_type
        ctypedef U mapped_type
        ctypedef pair[const T, U] value_type
        cppclass iterator:
            pair[T, U]& operator*()
            iterator operator++()
            iterator operator--()
            bint operator==(iterator)
            bint operator!=(iterator)
        cppclass reverse_iterator:
            pair[T, U]& operator*()
            iterator operator++()
            iterator operator--()
            bint operator==(reverse_iterator)
            bint operator!=(reverse_iterator)
        cppclass const_iterator(iterator):
            pass
        cppclass const_reverse_iterator(reverse_iterator):
            pass
        unordered_map() except +
        unordered_map(unordered_map&) except +
        U& operator[](T&)
        bint operator==(unordered_map&, unordered_map&)
        bint operator!=(unordered_map&, unordered_map&)
        bint operator<(unordered_map&, unordered_map&)
        bint operator>(unordered_map&, unordered_map&)
        bint operator<=(unordered_map&, unordered_map&)
        bint operator>=(unordered_map&, unordered_map&)
        U& at(T&)
        iterator begin()
        const_iterator const_begin "begin"()
        void clear()
        size_t count(T&)
        bint empty()
        iterator end()
        const_iterator const_end "end"()
        pair[iterator, iterator] equal_range(T&)
        iterator erase(iterator)
        iterator erase(iterator, iterator)
        size_t erase(T&)
        iterator find(T&)
        const_iterator const_find "find"(T&)
        pair[iterator, bint] insert(pair[T, U])  # XXX pair[T,U]&
        iterator insert(iterator, pair[T, U])  # XXX pair[T,U]&
        iterator lower_bound(T&)
        const_iterator const_lower_bound "lower_bound"(T&)
        size_t max_size()
        reverse_iterator rbegin()
        const_reverse_iterator const_rbegin "rbegin"()
        reverse_iterator rend()
        const_reverse_iterator const_rend "rend"()
        size_t size()
        void swap(unordered_map&)
        iterator upper_bound(T&)
        const_iterator const_upper_bound "upper_bound"(T&)
        void max_load_factor(float)
        float max_load_factor()

cdef extern from "bsl_unordered_set.h" namespace "bsl" nogil:
    cdef cppclass unordered_set[T]:
        ctypedef T value_type
        cppclass iterator:
            T& operator*()
            iterator operator++()
            iterator operator--()
            bint operator==(iterator)
            bint operator!=(iterator)
        cppclass reverse_iterator:
            T& operator*()
            iterator operator++()
            iterator operator--()
            bint operator==(reverse_iterator)
            bint operator!=(reverse_iterator)
        cppclass const_iterator(iterator):
            pass
        cppclass const_reverse_iterator(reverse_iterator):
            pass
        unordered_set() except +
        unordered_set(unordered_set&) except +
        bint operator==(unordered_set&, unordered_set&)
        bint operator!=(unordered_set&, unordered_set&)
        bint operator<(unordered_set&, unordered_set&)
        bint operator>(unordered_set&, unordered_set&)
        bint operator<=(unordered_set&, unordered_set&)
        bint operator>=(unordered_set&, unordered_set&)
        iterator begin()
        const_iterator const_begin "begin"()
        void clear()
        size_t count(T&)
        bint empty()
        iterator end()
        const_iterator const_end "end"()
        pair[iterator, iterator] equal_range(T&)
        void erase(iterator)
        void erase(iterator, iterator)
        size_t erase(T&)
        iterator find(T&)
        const_iterator const_find "find"(T&)
        pair[iterator, bint] insert(T&)
        iterator insert(iterator, T&)
        iterator lower_bound(T&)
        const_iterator const_lower_bound "lower_bound"(T&)
        size_t max_size()
        reverse_iterator rbegin()
        const_reverse_iterator const_rbegin "rbegin"()
        reverse_iterator rend()
        const_reverse_iterator const_rend "rend"()
        size_t size()
        void swap(unordered_set&)
        iterator upper_bound(T&)
        const_iterator const_upper_bound "upper_bound"(T&)

cdef extern from "bsl_optional.h" namespace "bsl" nogil:
    cdef cppclass optional[T]:
        optional()
        optional(T other)



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import sys
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Mapping
from typing import Optional
from typing import TYPE_CHECKING
from typing import Tuple
from typing import Type
from typing import Union
import weakref

from ._enums import AckStatus
from ._enums import PropertyType
from ._messages import Ack
from ._messages import Message
from ._messages import MessageHandle
from ._messages import create_ack
from ._messages import create_message
from ._messages import create_message_handle
from .session_events import InterfaceError
from .session_events import QueueEvent
from .session_events import QueueReopenFailed
from .session_events import QueueReopened
from .session_events import QueueResumeFailed
from .session_events import QueueResumed
from .session_events import QueueSuspendFailed
from .session_events import QueueSuspended
from .session_events import SessionEvent

if TYPE_CHECKING:
    # Safely perform circular references only during static type analysis
    from . import _ext  # pragma: no cover


def on_session_event(
    user_callback: Callable[[SessionEvent], None],
    event_type_mapping: Dict[int, Type[SessionEvent]],
    error_description: bytes,
    sdk_event: Optional[Tuple[int, bytes, int, bytes, str]] = None,
) -> None:
    if sdk_event is None:
        # This is a synthetically generated InterfaceError being produced in
        # response to input from the SDK that we can't handle.
        return user_callback(InterfaceError(error_description.decode()))

    # Otherwise, we're passing a bmqa::SessionEvent we've received to our user
    event_type, event_name, status_code, status_name, queue_uri = sdk_event
    event_cls = event_type_mapping.get(event_type, InterfaceError)

    # Prepare event message
    if event_cls is InterfaceError:
        msg = "Unexpected event type: %s" % event_name.decode()
    elif status_code != 0:
        msg = "%s%s%s (%d)" % (
            error_description.decode(),
            ": " if error_description else "",
            status_name.decode(),
            status_code,
        )
    else:
        msg = None

    # Create event
    if issubclass(event_cls, QueueEvent):
        failure_class_by_success_class = {
            QueueReopened: QueueReopenFailed,
            QueueResumed: QueueResumeFailed,
            QueueSuspended: QueueSuspendFailed,
        }

        if status_code != 0:
            event_cls = failure_class_by_success_class[event_cls]

        assert queue_uri
        event: SessionEvent = event_cls(queue_uri, msg)
    else:
        event = event_cls(msg)

    # Invoke user callback
    user_callback(event)


PropertiesAndTypesDictsType = Tuple[Dict[str, Union[int, bytes]], Dict[str, int]]


def on_message(
    user_callback: Callable[[Message, MessageHandle], None],
    ext_session_wr: weakref.ref[_ext.Session],
    property_type_to_py: Mapping[int, PropertyType],
    messages: Iterable[Tuple[bytes, bytes, bytes, PropertiesAndTypesDictsType]],
) -> None:
    ext_session = ext_session_wr()
    assert ext_session is not None, "ext.Session has been deleted"
    for data, guid, queue_uri, properties_tuple in messages:
        properties, property_types = properties_tuple
        property_types_py = {
            k: property_type_to_py[v] for k, v in property_types.items()
        }
        message = create_message(
            data, guid, queue_uri.decode(), properties, property_types_py
        )
        message_handle = create_message_handle(message, ext_session)
        user_callback(message, message_handle)

    del message_handle  # The message handle holds a reference to the extension session.
    if sys.getrefcount(ext_session) == 2:  # covered in a subprocess  # pragma: no cover
        # Dropping our reference to the extension session will drop its reference count
        # to 0, calling __dealloc__ and stop() from its own background thread.
        print(
            "Deadlock detected by blazingmq after calling %s; aborting process."
            % user_callback,
            file=sys.stderr,
        )
        try:
            import faulthandler

            faulthandler.dump_traceback()
        finally:
            # `faulthandler` only exists on 3; abort even if it doesn't exist or fails.
            sys.stderr.flush()
            os.abort()


def on_ack(
    ack_status_mapping: Dict[int, AckStatus],
    acks: Iterable[Tuple[int, bytes, Optional[bytes], bytes, Callable[[Ack], None]]],
) -> None:
    for status, status_description, guid, queue_uri, user_callback in acks:
        py_status = ack_status_mapping.get(status, AckStatus.UNRECOGNIZED)
        user_callback(
            create_ack(
                guid,
                py_status,
                status_description.decode(),
                queue_uri.decode(),
            )
        )


def on_message_create_interface_error(
    user_on_session_event: Callable[[SessionEvent], None],
    _: Any,
) -> None:
    error_description = "Messages received but no callback configured"
    user_on_session_event(InterfaceError(error_description))



# Copyright 2019-2023 Bloomberg Finance L.P.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

from blazingmq import Ack
from blazingmq import CompressionAlgorithmType
from blazingmq import Message
from blazingmq import MessageHandle
from blazingmq import PropertyType
from blazingmq import Timeouts
from blazingmq.session_events import SessionEvent

DEFAULT_MAX_UNCONFIRMED_MESSAGES: int = ...
DEFAULT_MAX_UNCONFIRMED_BYTES: int = ...
DEFAULT_CONSUMER_PRIORITY: int = ...
DEFAULT_SUSPENDS_ON_BAD_HOST_HEALTH: bool = ...

class FakeHostHealthMonitor:
    def __init__(self) -> None: ...
    def set_healthy(self) -> None: ...
    def set_unhealthy(self) -> None: ...

class Session:
    def __init__(
        self,
        on_session_event: Callable[[SessionEvent], None],
        *,
        on_message: Optional[Callable[[Message, MessageHandle], None]] = None,
        broker: bytes,
        message_compression_algorithm: CompressionAlgorithmType,
        num_processing_threads: Optional[int] = None,
        blob_buffer_size: Optional[int] = None,
        channel_high_watermark: Optional[int] = None,
        event_queue_watermarks: Optional[tuple[int, int]] = None,
        stats_dump_interval: Optional[int | float] = None,
        timeouts: Timeouts = Timeouts(),
        monitor_host_health: bool = False,
        fake_host_health_monitor: Optional[FakeHostHealthMonitor] = None,
    ) -> None: ...
    def stop(self) -> None: ...
    def open_queue_sync(
        self,
        queue_uri: bytes,
        *,
        read: bool,
        write: bool,
        consumer_priority: Optional[int] = None,
        max_unconfirmed_messages: Optional[int] = None,
        max_unconfirmed_bytes: Optional[int] = None,
        suspends_on_bad_host_health: Optional[bool] = None,
        timeout: Optional[float] = None,
    ) -> None: ...
    def close_queue_sync(
        self, queue_uri: bytes, *, timeout: Optional[float] = None
    ) -> None: ...
    def get_queue_options(self, queue_uri: bytes) -> Tuple[int, int, int]: ...
    def post(
        self,
        queue_uri: bytes,
        payload: bytes,
        *,
        properties: Optional[Dict[bytes, Tuple[Union[int, bytes], int]]] = None,
        on_ack: Optional[Callable[[Ack], None]] = None,
    ) -> None: ...
    def configure_queue_sync(
        self,
        queue_uri: bytes,
        *,
        max_unconfirmed_messages: Optional[int] = None,
        max_unconfirmed_bytes: Optional[int] = None,
        consumer_priority: Optional[int] = None,
        suspends_on_bad_host_health: Optional[bool] = None,
        timeout: Optional[float] = None,
    ) -> None: ...
    def confirm(self, message: Message) -> None: ...
    @property
    def monitor_host_health(self) -> bool: ...

PROPERTY_TYPES_FROM_PY_MAPPING: Dict[PropertyType, int]
